/**
 * The gMobilab driver was contributed by:
 * GNU/Linux version by Lucie Daubigney from Supelec Metz
 * Windows version by Andrea Villa from Universita' degli Studi di Milano
 * GNU/Linux - Windows drivers integration and revision by Enrico Calore from Universita' degli studi di Milano
 * Developed thanks to the support of g.tec medical engineering GmbH, Austria
 */

#include "ovasCDriverGTecGMobiLabPlus.h"
#include "ovasCConfigurationGTecGMobiLabPlus.h"
#include "../ovasCConfigurationBuilder.h"

#if defined TARGET_HAS_ThirdPartyGMobiLabPlusAPI

#include <openvibe-toolkit/ovtk_all.h>

#if defined OVAS_OS_Linux
   #include <system/Time.h>
   #include <cmath>
   #include <cstdlib>
   #include <cstdio>
#endif // defined OVAS_OS_Linux

#if defined OVAS_OS_Windows
//   #include <windows.h>
   #include <sstream>
   #include <iostream>
#endif // defined OVAS_OS_Windows

#include <cstring>
#include <iostream>

#define boolean OpenViBE::boolean

using namespace OpenViBEAcquisitionServer;
using namespace OpenViBE;
using namespace OpenViBE::Kernel;
using namespace std;

//___________________________________________________________________//
//                                                                   //

//constructor
CDriverGTecGMobiLabPlus::CDriverGTecGMobiLabPlus(IDriverContext& rDriverContext)
	:IDriver(rDriverContext)
	,m_pHeader(NULL)
	,m_pCallback(NULL)
	,m_ui32SampleCountPerSentBlock(0)
	,m_ui32DeviceIdentifier(uint32(-1))
	,m_pSample(NULL)
{
	m_rDriverContext.getLogManager() << LogLevel_Trace << "CDriverGTecGMobiLabPlus::CDriverGTecGMobiLabPlus\n";

	// Conversion factor to microVolt if multiplied by the channel sensitivity
	m_fConversionFactor = (2.f*5.f)/(4.f*65536.f); 
	
    m_bTestMode = false;
	
	m_pHeader = new CHeader();
	m_pHeader->setSamplingFrequency(256);
	m_pHeader->setChannelCount(8);

	m_oBuffer.pBuffer = NULL;
	m_oBuffer.size = 0;
	m_oBuffer.validPoints = 0;
	
}

CDriverGTecGMobiLabPlus::~CDriverGTecGMobiLabPlus(void)
{
	delete m_pHeader;
}

void CDriverGTecGMobiLabPlus::release(void)
{
	m_rDriverContext.getLogManager() << LogLevel_Trace << "CDriverGTecGMobiLabPlus::release\n";
	delete this;
}

const char* CDriverGTecGMobiLabPlus::getName(void)
{
	m_rDriverContext.getLogManager() << LogLevel_Trace << "CDriverGTecGMobiLabPlus::getName\n";
	return "gTec gMOBIlab+";
}

//___________________________________________________________________//
//                                                                   //

/*
 * configuration
 */

boolean CDriverGTecGMobiLabPlus::isConfigurable(void)
{
	m_rDriverContext.getLogManager() << LogLevel_Trace << "CDriverGTecGMobiLabPlus::isConfigurable\n";
	return true;
}

boolean CDriverGTecGMobiLabPlus::configure(void)
{
	m_rDriverContext.getLogManager() << LogLevel_Trace << "CDriverGTecGMobiLabPlus::configure\n";

	// We use CConfigurationGTecMobilabPlus configuration which is a class that inheritate from the CConfigurationBuilder class
	// The difference between these two classes is the addition of a member of class. This member allows to change the port where is connected the device.
	CConfigurationGTecGMobiLabPlus m_oConfiguration("../share/openvibe-applications/acquisition-server/interface-GTec-GMobiLabPlus.ui", m_ui32DeviceIdentifier, m_bTestMode);

	// We configure the Header with it...
	if(!m_oConfiguration.configure(*m_pHeader))
	{
		return false;
	}

	
	return true;
}

//___________________________________________________________________//
//                                                                   //

/*
 * initialisation
 */
boolean CDriverGTecGMobiLabPlus::initialize(const uint32 ui32SampleCountPerSentBlock, IDriverCallback& rCallback)
{
	m_rDriverContext.getLogManager() << LogLevel_Trace << "CDriverGTecGMobiLabPlus::initialize\n";

	if(m_rDriverContext.isConnected())
	{
		return false;
	}

	if(!m_pHeader->isChannelCountSet() || !m_pHeader->isSamplingFrequencySet())
	{
		m_rDriverContext.getLogManager() << LogLevel_Trace << "Either channel count or sampling frequency is not set\n";
		return false;
	}
	
	// Disable all analog channels... required ones will be enabled later	
	m_oAnalogIn.ain1 = false;
	m_oAnalogIn.ain2 = false;
	m_oAnalogIn.ain3 = false;
	m_oAnalogIn.ain4 = false;
	m_oAnalogIn.ain5 = false;
	m_oAnalogIn.ain6 = false;
	m_oAnalogIn.ain7 = false;
	m_oAnalogIn.ain8 = false;
	
	// This driver do not support DIO channels
	l_oDigitalInOut.dio1_enable = false;
	l_oDigitalInOut.dio2_enable = false;
	l_oDigitalInOut.dio3_enable = false;
	l_oDigitalInOut.dio4_enable = false;
	l_oDigitalInOut.dio5_enable = false;
	l_oDigitalInOut.dio6_enable = false;
	l_oDigitalInOut.dio7_enable = false;
	l_oDigitalInOut.dio8_enable = false;
	
		
#if defined OVAS_OS_Windows
   // Set default communication port if not configured: 
   if (m_ui32DeviceIdentifier == uint32(-1)) m_ui32DeviceIdentifier = 1;
   m_rDriverContext.getLogManager() << LogLevel_Trace << "Configured port name is: " << "COM" << m_ui32DeviceIdentifier << "\n";
#endif

#if defined OVAS_OS_Linux
   // Set default communication port if not configured: 
   if (m_ui32DeviceIdentifier == uint32(-1)) m_ui32DeviceIdentifier = 0;
   m_rDriverContext.getLogManager() << LogLevel_Trace << "Configured port name is: " << "/dev/rfcomm" << m_ui32DeviceIdentifier << "\n";
#endif
	
#if defined OVAS_OS_Windows
		
		LPSTR l_oPortName = new char[15]; // Port name is assumed to be shorter than 15 char.
		sprintf(l_oPortName, "\\\\.\\COM%d", m_ui32DeviceIdentifier);
		
		m_oDevice = ::GT_OpenDevice(l_oPortName);
		
		if(m_oDevice) {
			m_rDriverContext.getLogManager() << LogLevel_Trace << "Connected to port: "<< l_oPortName <<"\n";
		} else {
		   m_rDriverContext.getLogManager() << LogLevel_Trace << "Could not connect to port: "<< l_oPortName <<"\n";
		   delete [] m_oBuffer.pBuffer;
		   delete [] m_pSample;
		   m_pSample=NULL;
		   m_oBuffer.pBuffer=NULL;
		   //::CloseHandle(m_pEvent);
		   return false;
		}
		
		
	m_pEvent=::CreateEvent(NULL, FALSE, FALSE, NULL);
	if(!m_pEvent)
	{
		m_rDriverContext.getLogManager() << LogLevel_Trace << "Could not create m_pEvent \n";
		return false;
	}

#endif

#if defined OVAS_OS_Linux

    ostringstream l_oPortName;
	l_oPortName << "/dev/rfcomm" << m_ui32DeviceIdentifier; 
	m_oPortName = l_oPortName.str();

	m_oDevice=::GT_OpenDevice(m_oPortName.c_str());
	if(m_oDevice) {
	    m_rDriverContext.getLogManager() << LogLevel_Trace << "Connected to port: "<< l_oPortName <<"\n";
	} else {	
	    m_rDriverContext.getLogManager() << LogLevel_Trace << "Could not connect to port: " << m_oPortName << "\n";
		delete [] m_oBuffer.pBuffer;
		delete [] m_pSample;
		return false;
	}
#endif

	uint32 l_ui32ChannelCount = m_pHeader->getChannelCount();

	// analog exchanges allowed on the first "l_ui32CHannelCount" channels:
	for(uint32 i=1; i<=l_ui32ChannelCount; i++) 
	{
		CDriverGTecGMobiLabPlus::allowAnalogInputs(i);
	}

    // then buffer of type _BUFFER_ST built to store acquired samples.
	m_oBuffer.pBuffer=new short int[l_ui32ChannelCount]; //allocate enough space for the buffer m_oBuffer.pBuffer ; only one set of mesures is acquired (channel 1 to 8) in a row
	m_oBuffer.size=l_ui32ChannelCount*sizeof(short int);
	m_oBuffer.validPoints=0;
	
	// allocates enough space for m_pSample
	m_pSample = new float32[ ui32SampleCountPerSentBlock * l_ui32ChannelCount ];
	
	// if there is a problem while creating the two arrays
	if(!m_oBuffer.pBuffer || !m_pSample)
	{
		m_rDriverContext.getLogManager() << LogLevel_Trace << "Could not create buffers m_oBuffer.pBuffer or m_pSample \n";
		delete [] m_oBuffer.pBuffer;
		delete [] m_pSample;
		m_pSample=NULL;
		m_oBuffer.pBuffer=NULL;
		return false;
	}
		
	m_pCallback=&rCallback;
	m_ui32SampleCountPerSentBlock = ui32SampleCountPerSentBlock;

	return true;
}

boolean CDriverGTecGMobiLabPlus::uninitialize(void)
{
	m_rDriverContext.getLogManager() << LogLevel_Trace << "CDriverGTecGMobiLabPlus::uninitialize\n";

	if(!m_rDriverContext.isConnected()) { return false; }
	if(m_rDriverContext.isStarted()) { return false; }

	// uninitializes hardware here
	if (!::GT_CloseDevice(m_oDevice))
	{
		return false;
	}

	// frees memory
	delete [] m_pSample;
	delete [] m_oBuffer.pBuffer;

	m_pSample=NULL;
	m_oBuffer.pBuffer=NULL;
	m_pCallback=NULL;

	// uninitialisation of the analog channels : set valus to default ones
	m_oAnalogIn.ain1 = false;
	m_oAnalogIn.ain2 = false;
	m_oAnalogIn.ain3 = false;
	m_oAnalogIn.ain4 = false;
	m_oAnalogIn.ain5 = false;
	m_oAnalogIn.ain6 = false;
	m_oAnalogIn.ain7 = false;
	m_oAnalogIn.ain8 = false;
	
	::CloseHandle(m_pEvent);
	
	m_pEvent=NULL;

	return true;
}

const IHeader* CDriverGTecGMobiLabPlus::getHeader(void)
{
	return m_pHeader;
}

//___________________________________________________________________//
//                                                                   //

/*
 * acquisition
 */

boolean CDriverGTecGMobiLabPlus::start(void)
{
	m_rDriverContext.getLogManager() << LogLevel_Trace << "CDriverGTecGMobiLabPlus::start\n";

	if(!m_rDriverContext.isConnected()) { return false; }
	if(m_rDriverContext.isStarted()) { return false; }
	
	m_pOverlapped.hEvent = m_pEvent;
	m_pOverlapped.Offset = 0;
	m_pOverlapped.OffsetHigh = 0;

	// Reset deviceconfig structure
	char* hlp = (char*) &m_oDeviceConfig;
	for (unsigned int i = 0; i < sizeof(m_oDeviceConfig); i++) {
		hlp[i] = 0;
	}
	// Get device config
	if(!::GT_GetConfig(m_oDevice, &m_oDeviceConfig)) {
		m_rDriverContext.getLogManager() << LogLevel_Trace << "GT_GetConfig failed\n";
		return false;
	}
	// Print channels sensitivity for debugging:
	for (unsigned int i = 0; i < m_pHeader->getChannelCount(); i++) {
		m_rDriverContext.getLogManager() << LogLevel_Trace << "Channel " << i+1 << "sensitivity is set to: " << m_oDeviceConfig.channels[i].sensitivity << "\n";
	}
	
	// channel initialisation
	if(!::GT_InitChannels(m_oDevice, m_oAnalogIn, l_oDigitalInOut))	{
		m_rDriverContext.getLogManager() << LogLevel_Trace << "GT_InitChannels failed\n";
		return false;
	}
	
	// set or unset TestMode
	if(!::GT_SetTestmode(m_oDevice, m_bTestMode)) {
		m_rDriverContext.getLogManager() << LogLevel_Trace << "GT_SetTestMode status failed\n";
		return false;
	}

	// requests hardware to start sending data
	if(!::GT_StartAcquisition(m_oDevice))
	{
		m_rDriverContext.getLogManager() << LogLevel_Trace << "GT_StartAcquisition failed\n";
		return false;
	}
	return true;
}

boolean CDriverGTecGMobiLabPlus::loop(void)
{
	//m_rDriverContext.getLogManager() << LogLevel_Trace << "CDriverGTecGMobiLabPlus::loop  called \n";
			
    if(!m_rDriverContext.isConnected()) { return false; }
	if(!m_rDriverContext.isStarted()) { return true; }

	uint32 i, j;
	uint32 l_ui32ChannelCount=m_pHeader->getChannelCount();

	for(i = 0 ; i < m_ui32SampleCountPerSentBlock ; i++) {	
	
		DWORD dwBytesReceived = 0;
	
		#if defined OVAS_OS_Windows

		if (!::GT_GetData(m_oDevice, &m_oBuffer, &m_pOverlapped)) // receive samples from hardware (one per channel)
		{
			m_rDriverContext.getLogManager() << LogLevel_Trace << "GT_GetData failed\n";
			return false;
		}
	
		//WaitForSingleObject( m_pEvent, 1000);
		GetOverlappedResult( m_oDevice, &m_pOverlapped, &dwBytesReceived, true);
		m_oBuffer.validPoints = dwBytesReceived / 2; // short ints are 2 bytes long

		#endif

		#if defined OVAS_OS_Linux
		if (!::GT_GetData(m_oDevice, &m_oBuffer, &m_pOverlapped)) // receive samples from hardware (one per channel)
		{
			m_rDriverContext.getLogManager() << LogLevel_Trace << "GT_GetData failed\n";
			return false;
		}
		#endif 
		
		if (m_oBuffer.validPoints != l_ui32ChannelCount) {
	
			m_rDriverContext.getLogManager() << LogLevel_Trace << " Valid points are less than expected: " << m_oBuffer.validPoints << " instead of " << l_ui32ChannelCount << " \n";
		
			return false;
		
		} else {
				
			for ( j = 0; j < l_ui32ChannelCount; j++) {
				// The gmobilabplus order data in a different way than openvibe... we need to transpose the channel/sample matrix:
				// m_pSample expected by openvibe has to be ordered as:
				// [ [ch1][ch1][ch1][ch1] [ch2][ch2][ch2][ch2] [ch3][ch3][ch3][ch3] [ch4][ch4][ch4][ch4] ]
				// pBuffer given by the device is ordered as:
				// [ [ch1][ch2][ch3][ch4] [ch1][ch2][ch3][ch4] [ch1][ch2][ch3][ch4] [ch1][ch2][ch3][ch4] ]
				// The multi-purpose mobilab version has different sensitivity for the channels, thus different conversion factor
				// are needed to obtain a reading in microvolts for every channel.
				m_pSample[m_ui32SampleCountPerSentBlock*j+i] = OpenViBE::float32(m_oBuffer.pBuffer[j]) * m_oDeviceConfig.channels[j].sensitivity * m_fConversionFactor;
			}
		
			//m_rDriverContext.getLogManager() << LogLevel_Trace << "\n";
			//m_rDriverContext.getLogManager() << LogLevel_Trace << "m_oBuffer.size is: " << m_oBuffer.size << "\n";
			//m_rDriverContext.getLogManager() << LogLevel_Trace << "m_oBuffer.validPoints is: " << m_oBuffer.validPoints << "\n\n";
	
		}
	
		m_oBuffer.validPoints = 0;
	
	}
	
	// the buffer is full : it is send to the acquisition server
	m_pCallback->setSamples(m_pSample);
	m_rDriverContext.correctDriftSampleCount(m_rDriverContext.getSuggestedDriftCorrectionSampleCount());
	
	return true;
	
}

boolean CDriverGTecGMobiLabPlus::stop(void)
{
	m_rDriverContext.getLogManager() << LogLevel_Trace << "CDriverGTecGMobiLabPlus::stop\n";

	if(!m_rDriverContext.isConnected()) { return false; }
	if(!m_rDriverContext.isStarted()) { return false; }

	// requests the hardware to stop sending data
	if(!::GT_StopAcquisition(m_oDevice))
	{
		m_rDriverContext.getLogManager() << LogLevel_Trace << "GT_StopAcquisition failed\n";
		return false;
	}

	return true;
}

// this function allows exchanges of data on the "ui32ChannelIndex" channel
// function used to initialize the analog inputs according to the number "channelCount"
void CDriverGTecGMobiLabPlus::allowAnalogInputs(uint32 ui32ChannelIndex)
{
	switch(ui32ChannelIndex)
	{
		case 8: m_oAnalogIn.ain8 = true; break;
		case 7: m_oAnalogIn.ain7 = true; break;
		case 6: m_oAnalogIn.ain6 = true; break;
		case 5: m_oAnalogIn.ain5 = true; break;
		case 4: m_oAnalogIn.ain4 = true; break;
		case 3: m_oAnalogIn.ain3 = true; break;
		case 2: m_oAnalogIn.ain2 = true; break;
		case 1: m_oAnalogIn.ain1 = true; break;
		default:
			m_rDriverContext.getLogManager() << LogLevel_Trace << "Unexpected value " << ui32ChannelIndex << " in CDriverGTecGMobiLabPlus::allowAnalogInputs\n";
			break;
	}
}

#endif // TARGET_HAS_ThirdPartyGMobiLabPlusAPI


